from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Collection, Union

from light_curve.light_curve_ext import Extractor as _RustExtractor
from light_curve.light_curve_ext import _FeatureEvaluator as _RustBaseFeature

from ._base import BaseFeature
from .extractor import Extractor, _PyExtractor


@dataclass
class BaseMetaFeature(BaseFeature):
    features: Collection[Union[BaseFeature, _RustBaseFeature]] = ()
    extractor: Union[_RustExtractor, _PyExtractor] = field(init=False)

    def __post_init__(self):
        super().__post_init__()
        self.extractor = Extractor(*self.features)

    @abstractmethod
    def transform(self, t, m, sigma):
        """Must return temporarily sorted arrays (t, m, sigma)"""
        pass

    def _eval(self, t, m, sigma=None):
        raise NotImplementedError("_eval is missed for BaseMetaFeature")

    def _eval_and_fill(self, t, m, sigma, fill_value):
        t, m, sigma = self.transform(t, m, sigma)
        return self.extractor._eval_and_fill(t, m, sigma, fill_value)

    @property
    def size(self):
        return self.extractor.size



import numpy as np
import pytest

from light_curve.light_curve_py import MaximumSlope


def test_normalize():
    t = [1, 4, 3, 8]
    m = [2, 15, 4, 3]
    feature = MaximumSlope()
    actual = feature(t, m, None, sorted=False)
    desired = 11.0
    assert actual == desired


def test_fill_zero_division():
    t = [1, 1, 3, 4]
    feature = MaximumSlope()
    actual = feature(t, t, None, sorted=False, fill_value=1.0)
    desired = 1.0
    assert actual == desired


def test_fill_inf_values():
    t = [0, 1e-300]
    m = [0, 1e300]
    feature = MaximumSlope()
    actual = feature(t, m, None, sorted=True, fill_value=1.0)
    desired = 1.0
    assert actual == desired


def test_non_unique_values():
    t = [1, 1, 3, 4]
    feature = MaximumSlope()
    with pytest.raises(ValueError):
        feature(t, t, None, sorted=False)


def test_non_sorted_values():
    t = [2, 1, 3, 4]
    feature = MaximumSlope()
    with pytest.raises(ValueError):
        feature(t, t, None)



from .adnormal import *
from .amplitude import *
from .beyondnstd import *
from .bins import *
from .cusum import *
from .eta import *
from .etae import *
from .excvar import *
from .extractor import *
from .flux_n_not_det_before_fd import *
from .intpercrange import *
from .kurtosis import *
from .linfit import *
from .lintrend import *
from .magnitude_n_not_det_before_fd import *
from .magnpratio import *
from .maxslope import *
from .mean import *
from .meanvar import *
from .medabsdev import *
from .medbufrperc import *
from .median import *
from .otsusplit import *
from .pdiffmperc import *
from .percampl import *
from .redchi2 import *
from .skew import *
from .stdev import *
from .stetsonk import *
from .weightmean import *



import numpy as np
from scipy.stats import skew

from ._base import BaseFeature


class Skew(BaseFeature):
    def _eval(self, t, m, sigma=None):
        return skew(m, bias=False)

    @property
    def size(self):
        return 1


__all__ = ("Skew",)



import numpy as np

from ._base import BaseFeature


class StetsonK(BaseFeature):
    def _eval(self, t, m, sigma=None):
        m_mean = np.average(m, weights=np.power(sigma, -2))
        numerator = np.sum(np.abs((m - m_mean) / sigma))
        chisq = np.sum(((m - m_mean) / sigma) ** 2)
        return numerator / np.sqrt(len(m) * chisq)

    @property
    def size(self):
        return 1


__all__ = ("StetsonK",)



import numpy as np
from numpy.testing import assert_allclose
from scipy.optimize import curve_fit

from light_curve.light_curve_py import LinearFit


def test_linfit():
    def f(x, slope, c):
        return x * slope + c

    t = np.array([2.0, 4.0, 7.0, 10.0, 15.0, 28.0])
    m = np.array([1.0, 3.0, 6.0, 10.0, 14.0, 17.0])
    sigma = np.array([0.1, 0.1, 0.2, 1.3, 0.4, 1.5])
    n = len(t)

    feature = LinearFit()
    actual = feature(t, m, sigma)

    (slope, intercept), pcov = curve_fit(f, t, m, sigma=sigma, absolute_sigma=True)
    chisq = np.sum(((m - t * slope - intercept) / sigma) ** 2) / (n - 2)

    desired = (slope, np.sqrt(pcov[0][0]), chisq)
    assert_allclose(actual, desired)



import numpy as np
from scipy.stats import anderson

from ._base import BaseFeature


class AndersonDarlingNormal(BaseFeature):
    def _eval(self, t, m, sigma=None):
        n = len(m)
        return anderson(m).statistic * (1 + 4 / n - 25 / n**2)

    @property
    def size(self):
        return 1


__all__ = ("AndersonDarlingNormal",)



import numpy as np
from numpy.testing import assert_allclose

from light_curve.light_curve_py import Median


def test_median_1():
    n = 9
    m = np.arange(n)
    feature = Median()
    actual = feature(m, m, None)
    m = sorted(m)
    desired = m[n // 2]
    assert_allclose(actual, desired)


def test_median_2():
    n = 10
    m = np.arange(n)
    feature = Median()
    actual = feature(m, m, None)
    m = sorted(m)
    desired = (m[n // 2 - 1] + m[n // 2]) / 2
    assert_allclose(actual, desired)



import numpy as np

from ._base import BaseFeature


class ExcessVariance(BaseFeature):
    def _eval(self, t, m, sigma=None):
        m_mean = np.mean(m)
        d_mean = np.mean(np.power(sigma, 2))
        m_std = np.std(m, ddof=1)
        return (m_std**2 - d_mean) / m_mean**2

    @property
    def size(self):
        return 1


__all__ = ("ExcessVariance",)



import numpy as np
import pytest

from light_curve.light_curve_py import FluxNNotDetBeforeFd, MagnitudeNNotDetBeforeFd


def test_flux():
    feature = FluxNNotDetBeforeFd(10)
    m = np.array([1, 2, 10, 40, 50])
    t = [1, 2, 3, 4, 5]
    sigma = np.array([0.4, 0.2, 0.01, 0.03, 0.02])
    actual = feature(t, m, sigma)
    desired = 2
    assert actual == desired


def test_flux_strictly_fainter():
    feature = FluxNNotDetBeforeFd(10, strictly_fainter=True)
    m = np.array([1, 2, 2, 40, 50])
    t = [1, 2, 3, 4, 5]
    sigma = np.array([0.12, 0.2, 0.01, 0.03, 0.02])
    actual = feature(t, m, sigma)
    desired = 1
    assert actual == desired


def test_magnitude():
    feature = MagnitudeNNotDetBeforeFd(-1)
    m = np.array([10, 1, 2, 40, 50])
    sigma = [-1, -1, -1, 0.2, 0.03]
    t = [1, 2, 3, 4, 5]
    actual = feature(t, m, sigma)
    desired = 3
    assert actual == desired


def test_magnitude_strictly_fainter():
    feature = MagnitudeNNotDetBeforeFd(-1, strictly_fainter=True)
    m = np.array([10, 1, 2, 9, 50])
    sigma = [-1, -1, -1, 0.2, 0.03]
    t = [1, 2, 3, 4, 5]
    actual = feature(t, m, sigma)
    desired = 1
    assert actual == desired


def test_without_non_detections():
    feature1 = MagnitudeNNotDetBeforeFd(-1, strictly_fainter=True)
    feature2 = FluxNNotDetBeforeFd(-1, strictly_fainter=True)

    m = np.array([10, 1, 2, 9, 50])
    sigma = [0.1, 0.1, 0.1, 0.2, 0.03]
    t = [1, 2, 3, 4, 5]

    with pytest.raises(ValueError):
        feature1(t, m, sigma)
    with pytest.raises(ValueError):
        feature2(t, m, sigma)



from dataclasses import dataclass

import numpy as np
from scipy.stats.mstats import mquantiles

from ._base import BaseFeature


@dataclass()
class PercentDifferenceMagnitudePercentile(BaseFeature):
    p: float = 0.05

    def _eval(self, t, m, sigma=None):
        median = np.median(m)
        q1, q2 = mquantiles(m, [self.p, 1 - self.p], alphap=0.5, betap=0.5)
        return (q2 - q1) / median

    @property
    def size(self):
        return 1


__all__ = ("PercentDifferenceMagnitudePercentile",)



import inspect

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_array_equal

import light_curve.light_curve_ext as lc


def _feature_classes(module, exclude_parametric=True):
    for name, obj in inspect.getmembers(module):
        if name.startswith("_"):
            continue
        if inspect.ismodule(obj):
            yield from _feature_classes(obj)
        if not inspect.isclass(obj):
            continue
        if not issubclass(obj, lc._FeatureEvaluator):
            continue
        # Skip classes with non-trivial constructors
        if exclude_parametric:
            try:
                obj()
            except TypeError:
                continue
        yield obj


non_param_feature_classes = frozenset(_feature_classes(lc, True))
all_feature_classes = frozenset(_feature_classes(lc, False))


def gen_lc(n, rng=None):
    rng = np.random.default_rng(rng)

    t = np.sort(rng.normal(0, 1, n))
    m = t.copy()
    sigma = np.full_like(t, 0.1)

    return t, m, sigma


@pytest.mark.parametrize("cls", non_param_feature_classes)
def test_negative_strides(cls):
    t = np.linspace(1, 0, 20)[::-2]
    m = np.exp(t)[:]
    err = np.random.uniform(0.1, 0.2, t.shape)
    obj = cls()
    obj(t, m, err)


@pytest.mark.parametrize("cls", non_param_feature_classes)
def test_float32_vs_float64(cls):
    rng = np.random.default_rng(0)
    n = 128

    t, m, sigma = gen_lc(n, rng=rng)
    obj = cls()

    results = [
        obj(t.astype(dtype), m.astype(dtype), sigma.astype(dtype), sorted=True) for dtype in [np.float32, np.float64]
    ]
    assert_allclose(*results, rtol=1e-5, atol=1e-5)


@pytest.mark.parametrize("cls", non_param_feature_classes)
def test_many_vs_call(cls):
    rng = np.random.default_rng(0)
    n_obs = 128
    n_lc = 128

    lcs = [gen_lc(n_obs, rng=rng) for _ in range(n_lc)]
    obj = cls()

    call = np.stack([obj(*lc, sorted=True) for lc in lcs])
    many = obj.many(lcs, sorted=True, n_jobs=2)
    assert_array_equal(many, call)


def test_fill_value_not_enough_observations():
    n = 1
    t = np.linspace(0.0, 1.0, n)
    m = t.copy()
    fill_value = -100.0
    sigma = np.ones_like(t)
    feature = lc.Kurtosis()
    with pytest.raises(ValueError):
        feature(t, m, sigma, fill_value=None)
    assert_array_equal(feature(t, m, sigma, fill_value=fill_value), fill_value)


@pytest.mark.parametrize("cls", all_feature_classes)
def test_nonempty_docstring(cls):
    assert len(cls.__doc__) > 10


@pytest.mark.parametrize("cls", non_param_feature_classes)
def test_check_t(cls):
    n_obs = 128
    t, m, sigma = gen_lc(n_obs)
    t[0] = np.nan
    feature = cls()
    with pytest.raises(ValueError):
        feature(t, m, sigma, check=True)
    t[0] = np.inf
    with pytest.raises(ValueError):
        feature(t, m, sigma, check=True)


@pytest.mark.parametrize("cls", non_param_feature_classes)
def test_check_m(cls):
    n_obs = 128
    t, m, sigma = gen_lc(n_obs)
    m[0] = np.nan
    feature = cls()
    with pytest.raises(ValueError):
        feature(t, m, sigma, check=True)
    m[0] = np.inf
    with pytest.raises(ValueError):
        feature(t, m, sigma, check=True)


# We need evaluators which use sigma
@pytest.mark.parametrize("cls", (lc.ExcessVariance, lc.LinearFit, lc.ReducedChi2, lc.StetsonK, lc.WeightedMean))
def test_check_sigma(cls):
    n_obs = 128
    t, m, sigma = gen_lc(n_obs)
    sigma[0] = np.nan
    feature = cls()
    with pytest.raises(ValueError):
        feature(t, m, sigma, check=True)
    # infinite values are allowed for sigma
    sigma[0] = np.inf
    feature(t, m, sigma, check=True)



from dataclasses import dataclass

import numpy as np

from ._base import BaseFeature


@dataclass()
class BeyondNStd(BaseFeature):
    nstd: float = 1.0

    def _eval(self, t, m, sigma=None):
        mean = np.mean(m)
        std = np.std(m, ddof=1)
        return np.count_nonzero(np.abs(m - mean) > self.nstd * std) / len(m)

    @property
    def size(self):
        return 1


__all__ = ("BeyondNStd",)



import feets
import numpy as np
import pytest
from numpy.testing import assert_allclose
from scipy import stats
from scipy.optimize import curve_fit

import light_curve.light_curve_ext as lc_ext
import light_curve.light_curve_py as lc_py


class _Test:
    # Feature name must be updated in child classes
    name = None
    # Argument tuple for the feature constructor
    args = ()

    py_feature = None

    # Specify method for a naive implementation
    naive = None

    # Specify `feets` feature name
    feets_feature = None
    feets_extractor = None
    # Specify a `str` with a reason why skip this test
    feets_skip_test = False

    def setup_method(self):
        self.rust = getattr(lc_ext, self.name)(*self.args)
        try:
            self.py_feature = getattr(lc_py, self.name)(*self.args)
        except AttributeError:
            pass

        if self.feets_feature is not None:
            self.feets_extractor = feets.FeatureSpace(only=[self.feets_feature], data=["time", "magnitude", "error"])

    # Default values of `assert_allclose`
    rtol = 1e-7
    atol = 0

    # Default values for random light curve generation
    n_obs = 1000
    t_min = 0.0
    t_max = 1000.0
    m_min = 15.0
    m_max = 21.0
    sigma_min = 0.01
    sigma_max = 0.2

    add_to_all_features = True

    def generate_data(self):
        t = np.sort(np.random.uniform(self.t_min, self.t_max, self.n_obs))
        m = np.random.uniform(self.m_min, self.m_max, self.n_obs)
        sigma = np.random.uniform(self.sigma_min, self.sigma_max, self.n_obs)
        return t, m, sigma

    def test_feature_length(self):
        t, m, sigma = self.generate_data()
        result = self.rust(t, m, sigma, sorted=None)
        assert len(result) == len(self.rust.names) == len(self.rust.descriptions)

    def test_close_to_lc_py(self):
        if self.py_feature is None:
            pytest.skip("No matched light_curve_py class for the feature")
        t, m, sigma = self.generate_data()
        assert_allclose(self.rust(t, m, sigma), self.py_feature(t, m, sigma), rtol=self.rtol, atol=self.atol)

    def test_benchmark_rust(self, benchmark):
        t, m, sigma = self.generate_data()

        benchmark.group = str(type(self).__name__)
        benchmark(self.rust, t, m, sigma, sorted=True, check=False)

    def test_benchmark_lc_py(self, benchmark):
        if self.py_feature is None:
            pytest.skip("No matched light_curve_py class for the feature")

        t, m, sigma = self.generate_data()

        benchmark.group = str(type(self).__name__)
        benchmark(self.py_feature, t, m, sigma, sorted=True, check=False)

    def test_close_to_naive(self):
        if self.naive is None:
            pytest.skip("No naive implementation for the feature")

        t, m, sigma = self.generate_data()
        assert_allclose(self.rust(t, m, sigma), self.naive(t, m, sigma), rtol=self.rtol, atol=self.atol)

    def test_benchmark_naive(self, benchmark):
        if self.naive is None:
            pytest.skip("No naive implementation for the feature")

        t, m, sigma = self.generate_data()

        benchmark.group = type(self).__name__
        benchmark(self.naive, t, m, sigma)

    def feets(self, t, m, sigma):
        _, result = self.feets_extractor.extract(t, m, sigma)
        return result

    def test_close_to_feets(self):
        if self.feets_extractor is None:
            pytest.skip("No feets feature provided")
        if self.feets_skip_test:
            pytest.skip("feets is expected to be different from light_curve, reason: " + self.feets_skip_test)

        t, m, sigma = self.generate_data()
        assert_allclose(self.rust(t, m, sigma)[:1], self.feets(t, m, sigma)[:1], rtol=self.rtol, atol=self.atol)

    def test_benchmark_feets(self, benchmark):
        if self.feets_extractor is None:
            pytest.skip("No feets feature provided")

        t, m, sigma = self.generate_data()

        benchmark.group = type(self).__name__
        benchmark(self.feets, t, m, sigma)


class TestAmplitude(_Test):
    name = "Amplitude"

    def naive(self, t, m, sigma):
        return 0.5 * (np.max(m) - np.min(m))


class TestAndersonDarlingNormal(_Test):
    name = "AndersonDarlingNormal"

    feets_feature = "AndersonDarling"
    feets_skip_test = "feets uses biased statistics"

    def naive(self, t, m, sigma):
        return stats.anderson(m).statistic * (1.0 + 4.0 / m.size - 25.0 / m.size**2)


if lc_ext._built_with_gsl:

    class TestBazinFit(_Test):
        name = "BazinFit"
        args = ("lmsder", None, 20)
        rtol = 1e-4  # Precision used in the feature implementation

        add_to_all_features = False  # in All* random data is used

        @staticmethod
        def _model(t, a, b, t0, rise, fall):
            dt = t - t0
            return b + a * np.exp(-dt / fall) / (1.0 + np.exp(-dt / rise))

        def _params(self):
            a = 100
            c = 100
            t0 = 0.5 * (self.t_min + self.t_max)
            rise = 0.1 * (self.t_max - self.t_min)
            fall = 0.2 * (self.t_max - self.t_min)
            return a, c, t0, rise, fall

        # Random data yields to random results because target function has a lot of local minima
        # BTW, this test shouldn't use fixed random seed because the curve has good enough S/N to be fitted for any give
        # noise sample
        def generate_data(self):
            rng = np.random.default_rng(0)
            t = np.linspace(self.t_min, self.t_max, self.n_obs)
            model = self._model(t, *self._params())
            sigma = np.sqrt(model)
            m = model + sigma * rng.normal(size=self.n_obs)
            return t, m, sigma

        def naive(self, t, m, sigma):
            params, _cov = curve_fit(
                self._model,
                xdata=t,
                ydata=m,
                sigma=sigma,
                xtol=self.rtol,
                # We give really good parameters estimation!
                # p0=self._params(),
                # The same parameter estimations we use in Rust
                p0=(0.5 * np.ptp(m), np.min(m), t[np.argmax(m)], 0.5 * np.ptp(t), 0.5 * np.ptp(t)),
            )
            reduced_chi2 = np.sum(np.square((self._model(t, *params) - m) / sigma)) / (t.size - params.size)
            return_value = tuple(params) + (reduced_chi2,)
            return return_value


class TestBeyond1Std(_Test):
    nstd = 1.0

    name = "BeyondNStd"
    args = (nstd,)

    feets_feature = "Beyond1Std"
    feets_skip_test = "feets uses biased statistics"

    def naive(self, t, m, sigma):
        mean = np.mean(m)
        interval = self.nstd * np.std(m, ddof=1)
        return np.count_nonzero(np.abs(m - mean) > interval) / m.size


class TestCusum(_Test):
    name = "Cusum"

    feets_feature = "Rcs"
    feets_skip_test = "feets uses biased statistics"


class TestEta(_Test):
    name = "Eta"

    def naive(self, t, m, sigma):
        return np.sum(np.square(m[1:] - m[:-1])) / (np.var(m, ddof=0) * m.size)


class TestEtaE(_Test):
    name = "EtaE"

    feets_feature = "Eta_e"
    feets_skip_test = "feets fixed EtaE from the original paper in different way"

    def naive(self, t, m, sigma):
        return (
            np.sum(np.square((m[1:] - m[:-1]) / (t[1:] - t[:-1])))
            * (t[-1] - t[0]) ** 2
            / (np.var(m, ddof=0) * m.size * (m.size - 1) ** 2)
        )


class TestExcessVariance(_Test):
    name = "ExcessVariance"

    def naive(self, t, m, sigma):
        return (np.var(m, ddof=1) - np.mean(sigma**2)) / np.mean(m) ** 2


class TestInterPercentileRange(_Test):
    quantile = 0.25

    name = "InterPercentileRange"
    args = (quantile,)

    feets_feature = "Q31"
    feets_skip_test = "feets uses different quantile type"


class TestKurtosis(_Test):
    name = "Kurtosis"

    feets_feature = "SmallKurtosis"
    feets_skip_test = "feets uses equation for unbiased kurtosis, but put biased standard deviation there"

    def naive(self, t, m, sigma):
        return stats.kurtosis(m, fisher=True, bias=False)


class TestLinearTrend(_Test):
    name = "LinearTrend"

    feets_feature = "LinearTrend"

    def naive(self, t, m, sigma):
        (slope, _), ((slope_sigma2, _), _) = np.polyfit(t, m, deg=1, cov=True)
        sigma_noise = np.sqrt(np.polyfit(t, m, deg=1, full=True)[1][0] / (t.size - 2))
        return np.array([slope, np.sqrt(slope_sigma2), sigma_noise])


def generate_test_magnitude_percentile_ratio(quantile_numerator, quantile_denumerator, feets_feature):
    return type(
        f"TestMagnitudePercentageRatio{int(quantile_numerator * 100):d}",
        (_Test,),
        dict(
            quantile_numerator=quantile_numerator,
            quantile_denumerator=quantile_denumerator,
            name="MagnitudePercentageRatio",
            feets_skip_test="feets uses different quantile type",
        ),
    )


generate_test_magnitude_percentile_ratio(0.40, 0.05, "FluxPercentileRatioMid20")
generate_test_magnitude_percentile_ratio(0.25, 0.05, "FluxPercentileRatioMid50")
generate_test_magnitude_percentile_ratio(0.10, 0.05, "FluxPercentileRatioMid80")


class TestMaximumSlope(_Test):
    name = "MaximumSlope"

    feets_feature = "MaxSlope"

    def naive(self, t, m, sigma):
        return np.max(np.abs((m[1:] - m[:-1]) / (t[1:] - t[:-1])))


class TestMean(_Test):
    name = "Mean"

    feets_feature = "Mean"

    def naive(self, t, m, sigma):
        return np.mean(m)


class TestMeanVariance(_Test):
    name = "MeanVariance"

    feets_feature = "Meanvariance"
    feets_skip_test = "feets uses biased statistics"

    def naive(self, t, m, sigma):
        return np.std(m, ddof=1) / np.mean(m)


class TestMedian(_Test):
    name = "Median"

    def naive(self, t, m, sigma):
        return np.median(m)


class TestMedianAbsoluteDeviation(_Test):
    name = "MedianAbsoluteDeviation"

    feets_feature = "MedianAbsDev"


class TestMedianBufferRangePercentage(_Test):
    # feets says it uses 0.1 of amplitude (a half range between max and min),
    # but factually it uses 0.1 of full range between max and min
    quantile = 0.2

    name = "MedianBufferRangePercentage"
    args = (quantile,)

    feets_feature = "MedianBRP"


class TestPercentAmplitude(_Test):
    name = "PercentAmplitude"

    feets_feature = "PercentAmplitude"
    feets_skip_test = "feets divides value by median"

    def naive(self, t, m, sigma):
        median = np.median(m)
        return max(np.max(m) - median, median - np.min(m))


class TestPercentDifferenceMagnitudePercentile(_Test):
    quantile = 0.05

    name = "PercentDifferenceMagnitudePercentile"
    args = (quantile,)

    feets_feature = "PercentDifferenceFluxPercentile"
    feets_skip_test = "feets uses different quantile type"


class TestReducedChi2(_Test):
    name = "ReducedChi2"

    def naive(self, t, m, sigma):
        w = 1.0 / np.square(sigma)
        return np.sum(np.square(m - np.average(m, weights=w)) * w) / (m.size - 1)


class TestSkew(_Test):
    name = "Skew"

    feets_feature = "Skew"
    feets_skip_test = "feets uses biased statistics"

    def naive(self, t, m, sigma):
        return stats.skew(m, bias=False)


class TestStandardDeviation(_Test):
    name = "StandardDeviation"

    feets_feature = "Std"
    feets_skip_test = "feets uses biased statistics"

    def naive(self, t, m, sigma):
        return np.std(m, ddof=1)


class TestStetsonK(_Test):
    name = "StetsonK"

    feets_feature = "StetsonK"

    def naive(self, t, m, sigma):
        x = (m - np.average(m, weights=1.0 / sigma**2)) / sigma
        return np.sum(np.abs(x)) / np.sqrt(np.sum(np.square(x)) * m.size)


class TestWeightedMean(_Test):
    name = "WeightedMean"

    def naive(self, t, m, sigma):
        return np.average(m, weights=1.0 / sigma**2)


class TestAllPy(_Test):
    def setup_method(self):
        features = []
        py_features = []
        for cls in _Test.__subclasses__():
            if cls.name is None:
                continue

            try:
                py_features.append(getattr(lc_py, cls.name)(*cls.args))
            except AttributeError:
                continue
            features.append(getattr(lc_ext, cls.name)(*cls.args))
        self.rust = lc_ext.Extractor(*features)
        self.py_feature = lc_py.Extractor(*py_features)


class TestAllNaive(_Test):
    def setup_method(self):
        features = []
        self.naive_features = []
        for cls in _Test.__subclasses__():
            if cls.naive is None or cls.name is None:
                continue
            if not cls.add_to_all_features:
                continue
            if not cls.add_to_all_features:
                continue
            features.append(getattr(lc_ext, cls.name)(*cls.args))
            self.naive_features.append(cls().naive)
        self.rust = lc_ext.Extractor(*features)

    def naive(self, t, m, sigma):
        return np.concatenate([np.atleast_1d(f(t, m, sigma)) for f in self.naive_features])


class TestAllFeets(_Test):
    feets_skip_test = "skip for TestAllFeets"

    def setup_method(self):
        features = []
        feets_features = []
        for cls in _Test.__subclasses__():
            if cls.feets_feature is None or cls.name is None:
                continue
            if not cls.add_to_all_features:
                continue
            if not cls.add_to_all_features:
                continue
            features.append(getattr(lc_ext, cls.name)(*cls.args))
            feets_features.append(cls.feets_feature)
        self.rust = lc_ext.Extractor(*features)
        self.feets_extractor = feets.FeatureSpace(only=feets_features, data=["time", "magnitude", "error"])



import numpy as np

from light_curve.light_curve_ext import Eta, Mean
from light_curve.light_curve_ext import _FeatureEvaluator as _RustBaseFeature
from light_curve.light_curve_py import Amplitude, Extractor, OtsuSplit
from light_curve.light_curve_py.features._base import BaseFeature as _PythonBaseFeature


def test_extractor_all_python():
    t = np.linspace(0.0, 1.0, 50)
    perfect_m = 1e3 * t + 1e2
    err = np.sqrt(perfect_m)
    m = perfect_m + np.random.normal(0, err)

    otsusplit = OtsuSplit()
    amplitude = Amplitude()

    assert isinstance(otsusplit, _PythonBaseFeature)
    assert isinstance(amplitude, _PythonBaseFeature)

    extractor = Extractor(otsusplit, amplitude)
    actual = extractor(t, m, err)
    desired = np.concatenate([np.atleast_1d(otsusplit(t, m, err)), np.atleast_1d(amplitude(t, m, err))])
    assert np.all(actual == desired)


def test_extractor_all_rust():
    t = np.linspace(0.0, 1.0, 50)
    perfect_m = 1e3 * t + 1e2
    err = np.sqrt(perfect_m)
    m = perfect_m + np.random.normal(0, err)
    mean = Mean()
    eta = Eta()

    assert isinstance(mean, _RustBaseFeature)
    assert isinstance(eta, _RustBaseFeature)

    extractor = Extractor(mean, eta)
    actual = extractor(t, m, err)
    desired = np.concatenate([np.atleast_1d(mean(t, m, err)), np.atleast_1d(eta(t, m, err))])
    assert np.all(actual == desired)


def test_extractor_mixed_python_rust():
    t = np.linspace(0.0, 1.0, 50)
    perfect_m = 1e3 * t + 1e2
    err = np.sqrt(perfect_m)
    m = perfect_m + np.random.normal(0, err)
    eta = Eta()
    otsusplit = OtsuSplit()

    assert isinstance(eta, _RustBaseFeature)
    assert isinstance(otsusplit, _PythonBaseFeature)

    extractor = Extractor(eta, otsusplit)
    actual = extractor(t, m, err)
    desired = np.concatenate([np.atleast_1d(eta(t, m, err)), np.atleast_1d(otsusplit(t, m, err))])
    assert np.all(actual == desired)



from dataclasses import dataclass

import numpy as np

from ._base import BaseFeature


@dataclass()
class MedianBufferRangePercentage(BaseFeature):
    q: float = 0.1

    def _eval(self, t, m, sigma=None):
        median = np.median(m)
        return np.count_nonzero(np.abs(median - m) < self.q * (np.max(m) - np.min(m)) / 2) / len(m)

    @property
    def size(self):
        return 1


__all__ = ("MedianBufferRangePercentage",)



from dataclasses import dataclass
from typing import Tuple

import numpy as np

from ._base import BaseFeature


@dataclass()
class MagnitudeNNotDetBeforeFd(BaseFeature):
    """Number of non detections before the first detection for measurements of the magnitude.

    Feature use a user-defined value to mark non-detections: measurements with sigma equal to this value
    considered as non detections. strictly_fainter flag allows counting non-detections with a strictly larger
    upper limit than the first detection magnitude (there is no such feature in the original article).

    - Depends on: **magnitude**
    - Minimum number of observations: **2**
    - Number of features: **1**

    Attributes
    ----------
    sigma_non_detection : float
        Sigma value to mark the non detections values, may not be NaN.
    strictly_fainter : bool
        Flag to determine if to find non-detections with strictly larger upper limit than the first detection magnitude.

    P. Sánchez-Sáez et al 2021, [DOI:10.3847/1538-3881/abd5c1](https://doi.org/10.3847/1538-3881/abd5c1)
    """

    sigma_non_detection: float = np.Inf
    strictly_fainter: bool = False

    def _eval(self, t, m, sigma=None, sorted=None, fill_value=None):
        detections = np.argwhere(sigma != self.sigma_non_detection).flatten()

        if len(detections) == len(m):
            raise ValueError("There is no any non-detections")

        first_detection_idx = detections[0]

        if self.strictly_fainter:
            detection_m = m[first_detection_idx]
            # magnitude upper limits should be larger than the first detection
            non_detection_less = np.count_nonzero(m[:first_detection_idx] > detection_m)
            return non_detection_less

        return first_detection_idx

    @property
    def names(self) -> Tuple[str]:
        return ("magn_n_non_detections_before_fd",)

    @property
    def descriptions(self) -> Tuple[str]:
        return ("number of non detections before the first detection for magnitudes",)

    @property
    def size(self):
        return 1


__all__ = ("MagnitudeNNotDetBeforeFd",)



import numpy as np
from numpy.testing import assert_allclose

from light_curve.light_curve_py import Amplitude


def test_amplitude():
    n = 11
    feature = Amplitude()
    m = np.arange(n)
    desired = (n - 1) / 2
    actual = feature(m, m, None)
    assert_allclose(actual, desired)



import numpy as np

from ._base import BaseFeature


class MedianAbsoluteDeviation(BaseFeature):
    def _eval(self, t, m, sigma=None):
        return np.median((np.abs(m - np.median(m))))

    @property
    def size(self):
        return 1


__all__ = ("MedianAbsoluteDeviation",)



import numpy as np
from numpy.testing import assert_allclose

from light_curve.light_curve_py import MagnitudePercentageRatio


def test_magnpratio_1():
    m = np.arange(1.0, 11.0)
    feature = MagnitudePercentageRatio()
    feature2 = MagnitudePercentageRatio(0.4, 0.05)
    actual = feature(m, m, None)
    actual2 = feature2(m, m, None)
    desired = 0.222222222
    assert_allclose(actual, actual2, desired)


def test_magnpratio_2():
    m = np.arange(1.0, 11.0)
    feature = MagnitudePercentageRatio(0.5, 0.05)
    actual = feature(m, m, None)
    desired = 0.0
    assert_allclose(actual, desired)
